# This is spectacularly generated code by spectacular based on
# Qlik Cloud Services APIs

from __future__ import annotations

from dataclasses import asdict, dataclass
from typing import List, Literal

from ..auth import Auth, Config
from ..listable import ListableResource


@dataclass
class WebIntegration:
    """
    A web integration object.

    Attributes
    ----------
    created: str
      The time the web integration was created.
    createdBy: str
      The user that created the web integration.
    id: str
      The unique web integration identifier.
    lastUpdated: str
      The time the web integration was last updated.
    name: str
      The name of the web integration.
    tenantId: str
      The tenant that the web integration belongs to.
    validOrigins: list[str]
      The origins that are allowed to make requests to the tenant.
    """

    created: str = None
    createdBy: str = None
    id: str = None
    lastUpdated: str = None
    name: str = None
    tenantId: str = None
    validOrigins: list[str] = None

    def __init__(self_, **kvargs):
        if "created" in kvargs and kvargs["created"] is not None:
            self_.created = kvargs["created"]
        if "createdBy" in kvargs and kvargs["createdBy"] is not None:
            self_.createdBy = kvargs["createdBy"]
        if "id" in kvargs and kvargs["id"] is not None:
            self_.id = kvargs["id"]
        if "lastUpdated" in kvargs and kvargs["lastUpdated"] is not None:
            self_.lastUpdated = kvargs["lastUpdated"]
        if "name" in kvargs and kvargs["name"] is not None:
            self_.name = kvargs["name"]
        if "tenantId" in kvargs and kvargs["tenantId"] is not None:
            self_.tenantId = kvargs["tenantId"]
        if "validOrigins" in kvargs and kvargs["validOrigins"] is not None:
            self_.validOrigins = kvargs["validOrigins"]
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)

    def delete(self) -> None:
        """
        Delete web integration by ID
        Deletes a single web integration by ID.

        Parameters
        ----------
        """
        self.auth.rest(
            path="/web-integrations/{id}".replace("{id}", self.id),
            method="DELETE",
            params={},
            data=None,
        )

    def patch(self, data: WebIntegrationPatchSchema) -> None:
        """
        Update web integration by ID
        Updates a single web integration by ID.

        Parameters
        ----------
        data: WebIntegrationPatchSchema
          Properties that the user wants to update for the web integration.
        """
        if data is not None:
            try:
                data = asdict(data)
            except:
                data = data
        self.auth.rest(
            path="/web-integrations/{id}".replace("{id}", self.id),
            method="PATCH",
            params={},
            data=data,
        )


@dataclass
class WebIntegrationPatch:
    """
    A JSON Patch document as defined in http://tools.ietf.org/html/rfc6902.

    Attributes
    ----------
    op: Literal["replace"]
      The operation to be performed.
    path: Literal["/name", "/validOrigins"]
      A JSON Pointer.
    value: str
      New value to be used for this operation.
    """

    op: Literal["replace"] = None
    path: Literal["/name", "/validOrigins"] = None
    value: str = None

    def __init__(self_, **kvargs):
        if "op" in kvargs and kvargs["op"] is not None:
            self_.op = kvargs["op"]
        if "path" in kvargs and kvargs["path"] is not None:
            self_.path = kvargs["path"]
        if "value" in kvargs and kvargs["value"] is not None:
            self_.value = kvargs["value"]
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class WebIntegrationPatchSchema(List["WebIntegrationPatch"]):
    """

    Attributes
    ----------
    """

    def __init__(self_, elements):
        for e in elements:
            self_.append(WebIntegrationPatch(**e))


@dataclass
class WebIntegrationPost:
    """
    The creation of a web integration response.

    Attributes
    ----------
    created: str
      The time the web integration was created.
    createdBy: str
      The user that created the web integration.
    id: str
      The unique web integration identifier.
    lastUpdated: str
      The time the web integration was last updated.
    links: WebIntegrationPostLinks
      Pagination links
    name: str
      The name of the newly created web integration.
    tenantId: str
      The tenant that the web integration belongs to.
    validOrigins: list[str]
      The origins that are allowed to make requests to the tenant.
    """

    created: str = None
    createdBy: str = None
    id: str = None
    lastUpdated: str = None
    links: WebIntegrationPostLinks = None
    name: str = None
    tenantId: str = None
    validOrigins: list[str] = None

    def __init__(self_, **kvargs):
        if "created" in kvargs and kvargs["created"] is not None:
            self_.created = kvargs["created"]
        if "createdBy" in kvargs and kvargs["createdBy"] is not None:
            self_.createdBy = kvargs["createdBy"]
        if "id" in kvargs and kvargs["id"] is not None:
            self_.id = kvargs["id"]
        if "lastUpdated" in kvargs and kvargs["lastUpdated"] is not None:
            self_.lastUpdated = kvargs["lastUpdated"]
        if "links" in kvargs and kvargs["links"] is not None:
            if (
                type(kvargs["links"]).__name__
                == WebIntegrationPost.__annotations__["links"]
            ):
                self_.links = kvargs["links"]
            else:
                self_.links = WebIntegrationPostLinks(**kvargs["links"])
        if "name" in kvargs and kvargs["name"] is not None:
            self_.name = kvargs["name"]
        if "tenantId" in kvargs and kvargs["tenantId"] is not None:
            self_.tenantId = kvargs["tenantId"]
        if "validOrigins" in kvargs and kvargs["validOrigins"] is not None:
            self_.validOrigins = kvargs["validOrigins"]
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class WebIntegrationPostLinks:
    """
    Pagination links

    Attributes
    ----------
    self: WebIntegrationPostLinksSelf
      Link information for current page.
    """

    self: WebIntegrationPostLinksSelf = None

    def __init__(self_, **kvargs):
        if "self" in kvargs and kvargs["self"] is not None:
            if (
                type(kvargs["self"]).__name__
                == WebIntegrationPostLinks.__annotations__["self"]
            ):
                self_.self = kvargs["self"]
            else:
                self_.self = WebIntegrationPostLinksSelf(**kvargs["self"])
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class WebIntegrationPostLinksSelf:
    """
    Link information for current page.

    Attributes
    ----------
    href: str
      URL to the current page of records.
    """

    href: str = None

    def __init__(self_, **kvargs):
        if "href" in kvargs and kvargs["href"] is not None:
            self_.href = kvargs["href"]
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class WebIntegrationPostSchema:
    """

    Attributes
    ----------
    name: str
      The name of the web integration to create.
    validOrigins: list[str]
      The origins that are allowed to make requests to the tenant.
    """

    name: str = None
    validOrigins: list[str] = None

    def __init__(self_, **kvargs):
        if "name" in kvargs and kvargs["name"] is not None:
            self_.name = kvargs["name"]
        if "validOrigins" in kvargs and kvargs["validOrigins"] is not None:
            self_.validOrigins = kvargs["validOrigins"]
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class WebIntegrationsClass:
    """
    An array of web integration objects.

    Attributes
    ----------
    data: list[WebIntegration]
      Properties of web integrations in a given tenant.
    links: WebIntegrationsLinks
      Pagination links
    """

    data: list[WebIntegration] = None
    links: WebIntegrationsLinks = None

    def __init__(self_, **kvargs):
        if "data" in kvargs and kvargs["data"] is not None:
            if all(
                f"list[{type(e).__name__}]"
                == WebIntegrationsClass.__annotations__["data"]
                for e in kvargs["data"]
            ):
                self_.data = kvargs["data"]
            else:
                self_.data = [WebIntegration(**e) for e in kvargs["data"]]
        if "links" in kvargs and kvargs["links"] is not None:
            if (
                type(kvargs["links"]).__name__
                == WebIntegrationsClass.__annotations__["links"]
            ):
                self_.links = kvargs["links"]
            else:
                self_.links = WebIntegrationsLinks(**kvargs["links"])
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class WebIntegrationsLinks:
    """
    Pagination links

    Attributes
    ----------
    next: WebIntegrationsLinksNext
      Link information for next page.
    prev: WebIntegrationsLinksPrev
      Link information for previous page.
    self: WebIntegrationsLinksSelf
      Link information for current page.
    """

    next: WebIntegrationsLinksNext = None
    prev: WebIntegrationsLinksPrev = None
    self: WebIntegrationsLinksSelf = None

    def __init__(self_, **kvargs):
        if "next" in kvargs and kvargs["next"] is not None:
            if (
                type(kvargs["next"]).__name__
                == WebIntegrationsLinks.__annotations__["next"]
            ):
                self_.next = kvargs["next"]
            else:
                self_.next = WebIntegrationsLinksNext(**kvargs["next"])
        if "prev" in kvargs and kvargs["prev"] is not None:
            if (
                type(kvargs["prev"]).__name__
                == WebIntegrationsLinks.__annotations__["prev"]
            ):
                self_.prev = kvargs["prev"]
            else:
                self_.prev = WebIntegrationsLinksPrev(**kvargs["prev"])
        if "self" in kvargs and kvargs["self"] is not None:
            if (
                type(kvargs["self"]).__name__
                == WebIntegrationsLinks.__annotations__["self"]
            ):
                self_.self = kvargs["self"]
            else:
                self_.self = WebIntegrationsLinksSelf(**kvargs["self"])
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class WebIntegrationsLinksNext:
    """
    Link information for next page.

    Attributes
    ----------
    href: str
      URL to the next page of records.
    """

    href: str = None

    def __init__(self_, **kvargs):
        if "href" in kvargs and kvargs["href"] is not None:
            self_.href = kvargs["href"]
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class WebIntegrationsLinksPrev:
    """
    Link information for previous page.

    Attributes
    ----------
    href: str
      URL to the previous page of records.
    """

    href: str = None

    def __init__(self_, **kvargs):
        if "href" in kvargs and kvargs["href"] is not None:
            self_.href = kvargs["href"]
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class WebIntegrationsLinksSelf:
    """
    Link information for current page.

    Attributes
    ----------
    href: str
      URL to the current page of records.
    """

    href: str = None

    def __init__(self_, **kvargs):
        if "href" in kvargs and kvargs["href"] is not None:
            self_.href = kvargs["href"]
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


class WebIntegrations:
    def __init__(self, config: Config) -> None:
        self.config = config
        self.auth = Auth(config)

    def get(self, id: str) -> WebIntegration:
        """
        Get web integration by ID
        Retrieves a single web integration by ID.

        Parameters
        ----------
        id: str
          The ID of the web integration to retrieve.
        """
        response = self.auth.rest(
            path="/web-integrations/{id}".replace("{id}", id),
            method="GET",
            params={},
            data=None,
        )
        obj = WebIntegration(**response.json())
        obj.auth = self.auth
        return obj

    def get_web_integrations(
        self,
        endingBefore: str = None,
        limit: float = 10,
        sort: Literal["name", "+name", "-name"] = "+name",
        startingAfter: str = None,
        tenantId: str = None,
    ) -> ListableResource[WebIntegration]:
        """
        List web integrations
        Retrieves web integrations matching the query.

        Parameters
        ----------
        endingBefore: str = None
          The target web integration ID to start looking before for web integrations. Cannot be used in conjunction with startingAfter.
        limit: float = 10
          The number of web integration entries to retrieve.
        sort: Literal["name", "+name", "-name"] = "+name"
          The field to sort by. Prefix with +/- to indicate ascending/descending order.
        startingAfter: str = None
          The target web integration ID to start looking after for web integrations. Cannot be used in conjunction with endingBefore.
        tenantId: str = None
          The tenant ID to filter by.
        """
        query_params = {}
        if endingBefore is not None:
            query_params["endingBefore"] = endingBefore
        if limit is not None:
            query_params["limit"] = limit
        if sort is not None:
            query_params["sort"] = sort
        if startingAfter is not None:
            query_params["startingAfter"] = startingAfter
        if tenantId is not None:
            query_params["tenantId"] = tenantId
        response = self.auth.rest(
            path="/web-integrations",
            method="GET",
            params=query_params,
            data=None,
        )
        return ListableResource(
            response=response.json(),
            cls=WebIntegration,
            auth=self.auth,
            path="/web-integrations",
            query_params=query_params,
        )

    def create(self, data: WebIntegrationPostSchema) -> WebIntegrationPost:
        """
        Create web integration
        Creates a web integration.

        Parameters
        ----------
        data: WebIntegrationPostSchema
          Properties that the user wants to set for the web integration.
        """
        if data is not None:
            try:
                data = asdict(data)
            except:
                data = data
        response = self.auth.rest(
            path="/web-integrations",
            method="POST",
            params={},
            data=data,
        )
        obj = WebIntegrationPost(**response.json())
        obj.auth = self.auth
        return obj



from typing import Iterator, List, TypeVar
from urllib.parse import parse_qs, urlparse

from .auth import Auth

T = TypeVar("T")


def get_pagination_query_param(response: any):
    url_str: str = None
    try:
        url_str = response["links"]["next"]["href"]
    except:
        pass
    if not url_str:
        return False
    # should be "next" but for users it is "startingAfter"
    query_dict = parse_qs(urlparse(url_str).query)
    for k in ["next", "startingAfter"]:
        if k in query_dict:
            return k, query_dict[k][0]
    return False


class ListableResource(List[T]):
    pagination: Iterator[T]
    """
    Automatic paginating iterator for resources that supports it
    It handles fetching automatically the next set of data from the list

    Example:
    --------
    >>> items = Items(config).get_items()
    ... for item in items.pagination:
    ...     print_item(item)
    """

    def __init__(
        self,
        response,
        auth: Auth,
        cls=None,
        path="",
        method="GET",
        query_params={},
    ):
        has_data_response = isinstance(response, dict) and "data" in response
        is_paginated = has_data_response and "links" in response

        def create(objData: any):
            if cls is None:
                return objData
            obj = cls(**objData)
            obj.auth = auth
            return obj

        list.__init__(
            self,
            [
                create(elem)
                for elem in (response["data"] if has_data_response else response)
            ],
        )

        def gen():
            # Yield the elements from the first response
            for elem in self:
                yield elem

            # Make pagination requests if possible
            if is_paginated:
                pagination_params = query_params.copy()
                # use max limit to get the most elements per request
                pagination_params["limit"] = 100
                last_response = response
                done = False
                while not done:
                    pagination_query_param = get_pagination_query_param(
                        response=last_response
                    )
                    if pagination_query_param:
                        query_key, query_value = pagination_query_param
                        pagination_params[query_key] = query_value
                    else:
                        done = True
                    if not done:
                        raw_response = auth.rest(
                            method=method,
                            path=path,
                            data=None,
                            params=pagination_params,
                        )
                        last_response = raw_response.json()
                        for elem_data in last_response["data"]:
                            obj = create(elem_data)
                            # add the element to the list to make it accessible after the pagination too
                            self.append(obj)
                            yield obj

        self.pagination = gen()



import json
import struct
from dataclasses import dataclass
from functools import reduce
from threading import Condition, Lock, Thread
from typing import Callable, Dict, List, Literal
from urllib.parse import urlparse

import websocket

from ._version import __version__
from .config import Config
from .errors import EngineClosedException
from .interceptors.types import InterceptorHandler, Interceptors


def _get_json_data(x: any):
    """
    Get all the properties of a class as plain data.
    Similar to the asdict function but including dynamic properties.
    Excludes _session property
    """
    dict_data = None
    try:
        # get the __dict__ of the object
        # includes named and dynamic properties
        dict_data = x.__dict__
    except:
        if type(x) is dict:
            dict_data = x
    if dict_data is not None:
        dct = {}
        for k, v in dict_data.items():
            # exclude the private _session property
            if k != "_session":
                dct[k] = _get_json_data(v)
        return dct
    elif type(x) is list:
        return [_get_json_data(e) for e in x]
    return x


@dataclass
class RequestObject:
    # Identifier established by the initiator of the request
    id: int
    # name of the engine method.
    method: str
    # target of the method.
    handle: int
    # the parameters can be provided by name through an object or by position through an array
    params: any
    # version of JSON-RPC defaults to 2.0
    jsonrpc: str = "2.0"


@dataclass
class ResponseObject:
    # id of the backend object.
    id: int
    #  QIX type of the backend object. Can for example be "Doc" or "GenericVariable".
    type: str
    # Custom type of the backend object, if defined in qInfo.
    genericType: str
    # Handle of the backend object.
    handle: int
    # represents the returned value from engine
    result: dict


RequestInterceptor = Callable[[RequestObject], RequestObject]
""" RPC Request interceptor """
ResponseInterceptor = Callable[[dict], dict]
""" RPC Request interceptor """


class FailedToConnect(Exception):
    pass


class RpcSession:
    _interceptors: Interceptors[ResponseInterceptor, RequestInterceptor]

    _ws_url: str
    _headers: List[str]
    _recv_error: Exception
    _socket = None
    _watch_recv_thread = None

    def __init__(
        self, ws_url: str, headers: List[str] = None, interceptors: Interceptors = None
    ):
        if headers is None:
            headers = []
        if not ws_url:
            raise Exception("Empty url")
        self._headers = headers
        self._ws_url = ws_url
        self._interceptors = interceptors
        self.lock = Lock()
        self._listeners = {}
        self._id_listeners = {}
        self._recv_error = None

    def _emit(self, event, *args):
        if event in self._listeners:
            self._listeners[event](*args)
        if event == "closed":
            # call all the event closed listeners on all objects of session
            for handle in self._id_listeners:
                if "closed" in self._id_listeners[handle]:
                    for closed_listener in self._id_listeners[handle]["closed"]:
                        closed_listener(*args)
            # clear all listeners if session is closed
            self._listeners = {}
            self._id_listeners = {}

    def _emit_handle(self, handle, event, *args):
        if handle in self._id_listeners and event in self._id_listeners[handle]:
            for event_listener in self._id_listeners[handle][event]:
                event_listener(*args)
        if event == "closed":
            # clear all listeners if handle is closed
            self._clear_listeners(handle)

    def _emit_handles(self, handles, event, *args):
        for id in handles:
            self._emit_handle(id, event, *args)

    def _watch_recv(self):
        """
        _watch_recv watches for socket responses.
        Adds the response to _received.
        """

        while True:
            if not self.is_connected():
                return
            try:
                # Using code from the websocket-library documentation for receiving connection close status code
                # https://websocket-client.readthedocs.io/en/latest/examples.html#receiving-connection-close-status-codes
                resp_opcode, msg = self._socket.recv_data()
                if resp_opcode == 8 and len(msg) >= 2:
                    close_code = str(struct.unpack("!H", msg[0:2])[0])
                    close_msg = str(msg[2:])
                    self._socket = None
                    msg = False
                    # Assume that only close code >= 4000 is error
                    if int(close_code) >= 4000:
                        # Setting socket to None signals not connected which will close this thread
                        # The error will be thrown in the close function not in this thread because then
                        # it will not reach the caller
                        self._recv_error = EngineClosedException(
                            {"close_code": close_code, "close_msg": close_msg}
                        )
            except Exception as err:
                self._socket = None
                msg = False
                self._recv_error = err
            with self._received_added:
                if msg:
                    msg = json.loads(msg)
                    self._emit("traffic:*", "received", msg)
                    self._emit("traffic:received", msg)
                    # add response to _received and notify waiting
                    if "id" in msg:
                        self._received[msg["id"]] = msg
                        self._emit("message", msg)
                        self._emit_handle(msg["id"], "traffic:*", "received", msg)
                        self._emit_handle(msg["id"], "traffic:received", msg)
                        self._received_added.notify_all()
                    else:
                        self._emit(
                            "notification" if "params" in msg else "message", msg
                        )
                    if "change" in msg:
                        self._emit_handles(msg["change"], "changed")
                    if "close" in msg:
                        self._emit_handles(msg["close"], "closed")
                else:
                    # notify waiting receivers so that
                    # the not connected error can be raised
                    # if the error is raised from here then the
                    # wait_response will never finish
                    self._received_added.notify_all()

    def on(
        self,
        event_name: Literal[
            "traffic:*",
            "traffic:received",
            "traffic:sent",
            "closed",
            "opened",
            "notification",
            "message",
        ],
        listener: Callable[[any], None],
    ):
        """
        the on function handles the various states and communication events

        Parameters
        ----------
        event_name: str
            can be one of
            'traffic', 'traffic:received','traffic:sent',
            'closed', 'opened',
            'notification', 'message'

        Examples
        --------
        >>> from qlik_sdk import Auth, AuthType, Config
        ...
        ... rpc_session.on("closed", closed_listener)
        ... rpc_session.on("opened", opened_listener)
        ... with rpc_session.open() as rpc_client:
        ...     # ...
        """
        self._listeners[event_name] = listener

    def on_handle(self, handle, event_name, listener):
        if handle not in self._id_listeners:
            self._id_listeners[handle] = {event_name: [listener]}
        else:
            if event_name not in self._id_listeners[handle]:
                self._id_listeners[handle][event_name] = [listener]
            else:
                self._id_listeners[handle][event_name].append(listener)

    def _clear_listeners(self, handle):
        if handle in self._id_listeners:
            self._id_listeners[handle] = {}

    def open(self):
        """
        connect establishes a connection to provided url
        using the specified headers.

        If the client is already connected an exception will
        be raised.
        """
        if self.is_connected():
            raise Exception("Client already connected")
        socket = websocket.WebSocket()
        try:
            socket.connect(self._ws_url, header=self._headers, suppress_origin=True)
        except Exception as exc:
            raise FailedToConnect() from exc

        self._socket = socket
        self._received = {}
        self._id = -1
        self._received_added = Condition()

        self._watch_recv_thread = Thread(target=self._watch_recv)
        self._watch_recv_thread.start()
        self._emit("opened")
        return self

    def is_connected(self):
        """
        return connected state
        """
        return self._socket and self._socket.connected

    def close(self):
        """
        close closes the socket (if it's open).
        """

        if self.is_connected():
            self._socket.send_close()
            self._emit("closed")
        if self._watch_recv_thread is not None and self._watch_recv_thread.is_alive():
            self._watch_recv_thread.join()

        # Raise closed connection error here
        if self._recv_error:
            raise self._recv_error

    def __enter__(self):
        """
        __enter__ is called when client is used in a 'with' statement.
        """
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """
        __exit__ is called when the 'with' scope is exited. This will call
        the client's close method.
        """

        self.close()

    def send(self, method: str, handle: int, *args, **kwargs):
        """
        send is a thread-safe method that sends a websocket-message with the
        specified method, handle and parameters.
        The resulting response is returned.

        If the client isn't connected an exception is raised.

        Parameters
        ----------
        method: str
            engine method name for the request
        handle: int
            the associated handle
        args: tuple
            parameters by position in an array
        kwargs: dict
            parameters provided by name in an object
        """

        if not self.is_connected():
            raise Exception("Client not connected")

        self.lock.acquire()
        self._id += 1
        id_ = self._id
        self.lock.release()

        data = RequestObject(
            id=self._id,
            method=method,
            handle=handle,
            params=args if not kwargs else kwargs,
        )

        # send and wait respons
        data = reduce(lambda d, f: f(d), self._interceptors["request"].handlers, data)
        json_data = json.dumps(_get_json_data(data))
        self._socket.send(json_data)
        self._emit("traffic:*", "sent", data)
        self._emit("traffic:sent", data)
        self._emit_handle(handle, "traffic:*", "sent", data)
        self._emit_handle(handle, "traffic:sent", data)
        res = self._wait_response(id_)
        res["request_data"] = data
        res = reduce(lambda r, f: f(r), self._interceptors["response"].handlers, res)
        return_value = None
        if "result" in res:
            return_value = res["result"]
        elif "error" in res:
            raise Exception(res["error"]["message"])
        else:
            return_value = res
        return return_value

    def _wait_response(self, id_):
        """
        _wait_response waits (blocking) for a message with the specified id.
        Internal method that should only be called from send
        """

        with self._received_added:
            while id_ not in self._received:
                if not self.is_connected():
                    if self._recv_error:
                        raise self._recv_error
                    else:
                        raise Exception("not connected")
                self._received_added.wait()
            res = self._received[id_]
            del self._received[id_]
            return res


class RpcClient:
    __config: Config
    # property for storing the interceptors
    interceptors: Interceptors[ResponseInterceptor, RequestInterceptor]

    sessions: Dict[str, RpcSession] = {}

    def __init__(self, config) -> None:
        self.__config = config
        # initiating the interceptors
        self.interceptors = dict(
            request=InterceptorHandler[RequestInterceptor](),
            response=InterceptorHandler[ResponseInterceptor](),
        )

    def rpc(self, app_id: str) -> RpcSession:
        hostname = urlparse(self.__config.host).hostname
        ws_url = "wss://" + hostname.strip("/") + "/app/" + app_id
        version = __version__[1:]
        headers = [
            "Authorization: Bearer %s" % self.__config.api_key,
            f"User-Agent: qlik-sdk-python/{version}",
        ]
        if ws_url not in self.sessions:
            self.sessions[ws_url] = RpcSession(ws_url, headers, self.interceptors)
        return self.sessions[ws_url]


class RpcClientInstance:
    interceptors: Interceptors

    def __init__(self, rpcClient: RpcClient) -> None:
        self._rpcClient = rpcClient
        self.interceptors = rpcClient.interceptors

    def __call__(self, app_id: str) -> RpcSession:
        return self._rpcClient.rpc(app_id)



# This is spectacularly generated code by spectacular based on
# Qlik Cloud Services APIs

from __future__ import annotations

from dataclasses import asdict, dataclass
from enum import Enum
from typing import Literal

from ..auth import Auth, Config
from ..listable import ListableResource


class Analysis(Enum):
    Breakdown = "breakdown"
    ChangePoint = "changePoint"
    Comparison = "comparison"
    Contribution = "contribution"
    Correlation = "correlation"
    Fact = "fact"
    MutualInfo = "mutualInfo"
    Rank = "rank"
    Spike = "spike"
    Trend = "trend"
    Values = "values"


class AnalysisGroup(Enum):
    Anomaly = "anomaly"
    Brekadown = "brekadown"
    Comparison = "comparison"
    Correl = "correl"
    Fact = "fact"
    List = "list"
    MutualInfo = "mutualInfo"
    Rank = "rank"


class ChartType(Enum):
    Barchart = "barchart"
    Combochart = "combochart"
    Distributionplot = "distributionplot"
    Kpi = "kpi"
    Linechart = "linechart"
    Map = "map"
    Scatterplot = "scatterplot"
    Table = "table"


@dataclass
class AppMetadata:
    """
    Metadata for app

    Attributes
    ----------
    id: str
    last_reload_date: str
    limited_access: bool
    name: str
    space_id: str
    space_name: str
    space_type: str
    """

    id: str = None
    last_reload_date: str = None
    limited_access: bool = None
    name: str = None
    space_id: str = None
    space_name: str = None
    space_type: str = None

    def __init__(self_, **kvargs):
        if "id" in kvargs and kvargs["id"] is not None:
            self_.id = kvargs["id"]
        if "last_reload_date" in kvargs and kvargs["last_reload_date"] is not None:
            self_.last_reload_date = kvargs["last_reload_date"]
        if "limited_access" in kvargs and kvargs["limited_access"] is not None:
            self_.limited_access = kvargs["limited_access"]
        if "name" in kvargs and kvargs["name"] is not None:
            self_.name = kvargs["name"]
        if "space_id" in kvargs and kvargs["space_id"] is not None:
            self_.space_id = kvargs["space_id"]
        if "space_name" in kvargs and kvargs["space_name"] is not None:
            self_.space_name = kvargs["space_name"]
        if "space_type" in kvargs and kvargs["space_type"] is not None:
            self_.space_type = kvargs["space_type"]
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class ConversationalResponsePart:
    """

    Attributes
    ----------
    errorMessage: str
    followupSentence: str
    imageUrl: str
    infoType: str
    infoValues: list[any]
    narrative: NarrativeResponse
    renderVisualization: RenderVisualization
    sentence: ConversationalResponsePartSentence
    type: str
    """

    errorMessage: str = None
    followupSentence: str = None
    imageUrl: str = None
    infoType: str = None
    infoValues: list[any] = None
    narrative: NarrativeResponse = None
    renderVisualization: RenderVisualization = None
    sentence: ConversationalResponsePartSentence = None
    type: str = None

    def __init__(self_, **kvargs):
        if "errorMessage" in kvargs and kvargs["errorMessage"] is not None:
            self_.errorMessage = kvargs["errorMessage"]
        if "followupSentence" in kvargs and kvargs["followupSentence"] is not None:
            self_.followupSentence = kvargs["followupSentence"]
        if "imageUrl" in kvargs and kvargs["imageUrl"] is not None:
            self_.imageUrl = kvargs["imageUrl"]
        if "infoType" in kvargs and kvargs["infoType"] is not None:
            self_.infoType = kvargs["infoType"]
        if "infoValues" in kvargs and kvargs["infoValues"] is not None:
            self_.infoValues = kvargs["infoValues"]
        if "narrative" in kvargs and kvargs["narrative"] is not None:
            if (
                type(kvargs["narrative"]).__name__
                == ConversationalResponsePart.__annotations__["narrative"]
            ):
                self_.narrative = kvargs["narrative"]
            else:
                self_.narrative = NarrativeResponse(**kvargs["narrative"])
        if (
            "renderVisualization" in kvargs
            and kvargs["renderVisualization"] is not None
        ):
            if (
                type(kvargs["renderVisualization"]).__name__
                == ConversationalResponsePart.__annotations__["renderVisualization"]
            ):
                self_.renderVisualization = kvargs["renderVisualization"]
            else:
                self_.renderVisualization = RenderVisualization(
                    **kvargs["renderVisualization"]
                )
        if "sentence" in kvargs and kvargs["sentence"] is not None:
            if (
                type(kvargs["sentence"]).__name__
                == ConversationalResponsePart.__annotations__["sentence"]
            ):
                self_.sentence = kvargs["sentence"]
            else:
                self_.sentence = ConversationalResponsePartSentence(
                    **kvargs["sentence"]
                )
        if "type" in kvargs and kvargs["type"] is not None:
            self_.type = kvargs["type"]
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class ConversationalResponsePartSentence:
    """

    Attributes
    ----------
    text: str
    """

    text: str = None

    def __init__(self_, **kvargs):
        if "text" in kvargs and kvargs["text"] is not None:
            self_.text = kvargs["text"]
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class Error:
    """
    An error object.

    Attributes
    ----------
    code: str
      The error code.
    detail: str
      A human-readable explanation specific to this occurrence of the problem.
    meta: object
      Additional properties relating to the error.
    source: ErrorSource
      References to the source of the error.
    title: str
      Summary of the problem.
    """

    code: str = None
    detail: str = None
    meta: object = None
    source: ErrorSource = None
    title: str = None

    def __init__(self_, **kvargs):
        if "code" in kvargs and kvargs["code"] is not None:
            self_.code = kvargs["code"]
        if "detail" in kvargs and kvargs["detail"] is not None:
            self_.detail = kvargs["detail"]
        if "meta" in kvargs and kvargs["meta"] is not None:
            self_.meta = kvargs["meta"]
        if "source" in kvargs and kvargs["source"] is not None:
            if type(kvargs["source"]).__name__ == Error.__annotations__["source"]:
                self_.source = kvargs["source"]
            else:
                self_.source = ErrorSource(**kvargs["source"])
        if "title" in kvargs and kvargs["title"] is not None:
            self_.title = kvargs["title"]
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class ErrorSource:
    """
    References to the source of the error.

    Attributes
    ----------
    parameter: str
      The URI query parameter that caused the error.
    pointer: str
      A JSON Pointer to the property that caused the error.
    """

    parameter: str = None
    pointer: str = None

    def __init__(self_, **kvargs):
        if "parameter" in kvargs and kvargs["parameter"] is not None:
            self_.parameter = kvargs["parameter"]
        if "pointer" in kvargs and kvargs["pointer"] is not None:
            self_.pointer = kvargs["pointer"]
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class Href:
    """

    Attributes
    ----------
    href: str
    """

    href: str = None

    def __init__(self_, **kvargs):
        if "href" in kvargs and kvargs["href"] is not None:
            self_.href = kvargs["href"]
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class Links:
    """

    Attributes
    ----------
    next: Href
    prev: Href
    self: Href
    """

    next: Href = None
    prev: Href = None
    self: Href = None

    def __init__(self_, **kvargs):
        if "next" in kvargs and kvargs["next"] is not None:
            if type(kvargs["next"]).__name__ == Links.__annotations__["next"]:
                self_.next = kvargs["next"]
            else:
                self_.next = Href(**kvargs["next"])
        if "prev" in kvargs and kvargs["prev"] is not None:
            if type(kvargs["prev"]).__name__ == Links.__annotations__["prev"]:
                self_.prev = kvargs["prev"]
            else:
                self_.prev = Href(**kvargs["prev"])
        if "self" in kvargs and kvargs["self"] is not None:
            if type(kvargs["self"]).__name__ == Links.__annotations__["self"]:
                self_.self = kvargs["self"]
            else:
                self_.self = Href(**kvargs["self"])
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class NLMetricRecommendation:
    """
    Visualisation recommendation specs for the query

    Attributes
    ----------
    analysis: Literal["breakdown", "changePoint", "comparison", "contribution", "correlation", "fact", "mutualInfo", "rank", "spike", "trend", "values"]
    analysisGroup: Literal["anomaly", "brekadown", "comparison", "correl", "fact", "list", "mutualInfo", "rank"]
    chartType: Literal["barchart", "combochart", "distributionplot", "kpi", "linechart", "map", "scatterplot", "table"]
      Chart type given to current recommendation
    dims: list[str]
      Dimension(s) considered for recommendation
    msrs: list[str]
      Measure(s) considered for recommendation
    relevance: float
    """

    analysis: Analysis = None
    analysisGroup: AnalysisGroup = None
    chartType: ChartType = None
    dims: list[str] = None
    msrs: list[str] = None
    relevance: float = None

    def __init__(self_, **kvargs):
        if "analysis" in kvargs and kvargs["analysis"] is not None:
            if (
                type(kvargs["analysis"]).__name__
                == NLMetricRecommendation.__annotations__["analysis"]
            ):
                self_.analysis = kvargs["analysis"]
            else:
                self_.analysis = Analysis(kvargs["analysis"])
        if "analysisGroup" in kvargs and kvargs["analysisGroup"] is not None:
            if (
                type(kvargs["analysisGroup"]).__name__
                == NLMetricRecommendation.__annotations__["analysisGroup"]
            ):
                self_.analysisGroup = kvargs["analysisGroup"]
            else:
                self_.analysisGroup = AnalysisGroup(kvargs["analysisGroup"])
        if "chartType" in kvargs and kvargs["chartType"] is not None:
            if (
                type(kvargs["chartType"]).__name__
                == NLMetricRecommendation.__annotations__["chartType"]
            ):
                self_.chartType = kvargs["chartType"]
            else:
                self_.chartType = ChartType(kvargs["chartType"])
        if "dims" in kvargs and kvargs["dims"] is not None:
            self_.dims = kvargs["dims"]
        if "msrs" in kvargs and kvargs["msrs"] is not None:
            self_.msrs = kvargs["msrs"]
        if "relevance" in kvargs and kvargs["relevance"] is not None:
            self_.relevance = kvargs["relevance"]
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class NLMetricRecord:
    """

    Attributes
    ----------
    appId: str
      Qlik sense app id that is being used to answer the question
    appName: str
      Qlik sense app name that is being used to answer the question
    apps: list[AppMetadata]
    channelId: str
      Source from which conversation is happening
    chartType: str
      Chart type for given query. For insight advisor it would be 'native' and for insight advisor chat, it could be 'static' or 'responsive'
    createdAt: str
      Record created date
    createdBy: str
      Qlik sense user id who is interacting with insight advisor or insight advisor chat or third party api
    feedback: list[RecFeedback]
    id: str
      Unique record id stored in database
    isContextualQuery: bool
      Boolean value indicates whether given query is contextual or not. It would be false for insight advisor
    lang: str
      language selected for query from insight advisor or insight advisor chat or third party api
    nluInfo: list[PartialNluInfo]
    queryError: bool
    queryOrigin: Literal["askQuestion", "iaAnalysis", "iaAssetsPanel"]
      Refers to source from where narrative request is called
    queryText: str
      Query asked by user in insight advisor or insight advisor or third party api
    queryType: Literal["appList", "appSuggested", "dimensionList", "exploreThisFurther", "followup", "greetings", "measureList", "query", "sampleQuestion"]
      Nature of query being asked during the conversation e.g. query, applist, measurelist, dimensionlist
    questionId: str
      Unique id assigned to user query
    recommendations: list[NLMetricRecommendation]
    responses: NLResponses
      Provides info what was included in response for given query
    stopWords: list[str]
      Tokens from question parsed which are ignored
    tenantId: str
      Qlik sense tenant Id
    unmatchedEntities: list[str]
      Tokens parsed as entities but not matched with app's field/dimension/measure
    updatedAt: str
      Record modified date
    version: str
      Version of the metric model
    """

    appId: str = None
    appName: str = None
    apps: list[AppMetadata] = None
    channelId: str = None
    chartType: str = None
    createdAt: str = None
    createdBy: str = None
    feedback: list[RecFeedback] = None
    id: str = None
    isContextualQuery: bool = None
    lang: str = None
    nluInfo: list[PartialNluInfo] = None
    queryError: bool = None
    queryOrigin: Literal["askQuestion", "iaAnalysis", "iaAssetsPanel"] = "askQuestion"
    queryText: str = None
    queryType: Literal[
        "appList",
        "appSuggested",
        "dimensionList",
        "exploreThisFurther",
        "followup",
        "greetings",
        "measureList",
        "query",
        "sampleQuestion",
    ] = None
    questionId: str = None
    recommendations: list[NLMetricRecommendation] = None
    responses: NLResponses = None
    stopWords: list[str] = None
    tenantId: str = None
    unmatchedEntities: list[str] = None
    updatedAt: str = None
    version: str = None

    def __init__(self_, **kvargs):
        if "appId" in kvargs and kvargs["appId"] is not None:
            self_.appId = kvargs["appId"]
        if "appName" in kvargs and kvargs["appName"] is not None:
            self_.appName = kvargs["appName"]
        if "apps" in kvargs and kvargs["apps"] is not None:
            if all(
                f"list[{type(e).__name__}]" == NLMetricRecord.__annotations__["apps"]
                for e in kvargs["apps"]
            ):
                self_.apps = kvargs["apps"]
            else:
                self_.apps = [AppMetadata(**e) for e in kvargs["apps"]]
        if "channelId" in kvargs and kvargs["channelId"] is not None:
            self_.channelId = kvargs["channelId"]
        if "chartType" in kvargs and kvargs["chartType"] is not None:
            self_.chartType = kvargs["chartType"]
        if "createdAt" in kvargs and kvargs["createdAt"] is not None:
            self_.createdAt = kvargs["createdAt"]
        if "createdBy" in kvargs and kvargs["createdBy"] is not None:
            self_.createdBy = kvargs["createdBy"]
        if "feedback" in kvargs and kvargs["feedback"] is not None:
            if all(
                f"list[{type(e).__name__}]"
                == NLMetricRecord.__annotations__["feedback"]
                for e in kvargs["feedback"]
            ):
                self_.feedback = kvargs["feedback"]
            else:
                self_.feedback = [RecFeedback(**e) for e in kvargs["feedback"]]
        if "id" in kvargs and kvargs["id"] is not None:
            self_.id = kvargs["id"]
        if "isContextualQuery" in kvargs and kvargs["isContextualQuery"] is not None:
            self_.isContextualQuery = kvargs["isContextualQuery"]
        if "lang" in kvargs and kvargs["lang"] is not None:
            self_.lang = kvargs["lang"]
        if "nluInfo" in kvargs and kvargs["nluInfo"] is not None:
            if all(
                f"list[{type(e).__name__}]" == NLMetricRecord.__annotations__["nluInfo"]
                for e in kvargs["nluInfo"]
            ):
                self_.nluInfo = kvargs["nluInfo"]
            else:
                self_.nluInfo = [PartialNluInfo(**e) for e in kvargs["nluInfo"]]
        if "queryError" in kvargs and kvargs["queryError"] is not None:
            self_.queryError = kvargs["queryError"]
        if "queryOrigin" in kvargs and kvargs["queryOrigin"] is not None:
            self_.queryOrigin = kvargs["queryOrigin"]
        if "queryText" in kvargs and kvargs["queryText"] is not None:
            self_.queryText = kvargs["queryText"]
        if "queryType" in kvargs and kvargs["queryType"] is not None:
            self_.queryType = kvargs["queryType"]
        if "questionId" in kvargs and kvargs["questionId"] is not None:
            self_.questionId = kvargs["questionId"]
        if "recommendations" in kvargs and kvargs["recommendations"] is not None:
            if all(
                f"list[{type(e).__name__}]"
                == NLMetricRecord.__annotations__["recommendations"]
                for e in kvargs["recommendations"]
            ):
                self_.recommendations = kvargs["recommendations"]
            else:
                self_.recommendations = [
                    NLMetricRecommendation(**e) for e in kvargs["recommendations"]
                ]
        if "responses" in kvargs and kvargs["responses"] is not None:
            if (
                type(kvargs["responses"]).__name__
                == NLMetricRecord.__annotations__["responses"]
            ):
                self_.responses = kvargs["responses"]
            else:
                self_.responses = NLResponses(**kvargs["responses"])
        if "stopWords" in kvargs and kvargs["stopWords"] is not None:
            self_.stopWords = kvargs["stopWords"]
        if "tenantId" in kvargs and kvargs["tenantId"] is not None:
            self_.tenantId = kvargs["tenantId"]
        if "unmatchedEntities" in kvargs and kvargs["unmatchedEntities"] is not None:
            self_.unmatchedEntities = kvargs["unmatchedEntities"]
        if "updatedAt" in kvargs and kvargs["updatedAt"] is not None:
            self_.updatedAt = kvargs["updatedAt"]
        if "version" in kvargs and kvargs["version"] is not None:
            self_.version = kvargs["version"]
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class NLMetricsRequest:
    """

    Attributes
    ----------
    filter: str
      The advanced filtering to use for the query. Refer to RFC 7644 https://www.rfc-editor.org/rfc/rfc7644section-3.4.2.2 for the syntax.:

      Filter on createdAt and updatedAt fields are encouraged and support `eq`, `ne`, `gt`, `ge`, `lt`, `le` comparison operators along with `and` and `or` logical operators.

      Filter on tenantId field is not supported.

      `co`, `sw` and `ew` operators are not supported.

      Examples:
      ```
      appId eq "appId1"
      ```
      ```
      (appId eq "appId1" or appId eq "appId2")
      ```
      ```
      (appId eq "appId1" or appId eq "appId2") and (createdAt gt "2022-08-03T00:00:00.000Z" and createdAt lt "2022-08-04T00:00:00.000Z")
      ```

      ```
      (appId eq "appId1") and (createdAt ge "2022-08-03T00:00:00.000Z")
      ```

      ```
      (appId eq "appId1") and (createdAt le "2022-08-23:59:59.000Z")
      ```

      ```
      (appId eq "appId1") and (questionId eq "12345")
      ```

    """

    filter: str = None

    def __init__(self_, **kvargs):
        if "filter" in kvargs and kvargs["filter"] is not None:
            self_.filter = kvargs["filter"]
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class NLMetricsResponse:
    """

    Attributes
    ----------
    data: list[NLMetricRecord]
    links: Links
    meta: NLMetricsResponseMeta
    """

    data: list[NLMetricRecord] = None
    links: Links = None
    meta: NLMetricsResponseMeta = None

    def __init__(self_, **kvargs):
        if "data" in kvargs and kvargs["data"] is not None:
            if all(
                f"list[{type(e).__name__}]" == NLMetricsResponse.__annotations__["data"]
                for e in kvargs["data"]
            ):
                self_.data = kvargs["data"]
            else:
                self_.data = [NLMetricRecord(**e) for e in kvargs["data"]]
        if "links" in kvargs and kvargs["links"] is not None:
            if (
                type(kvargs["links"]).__name__
                == NLMetricsResponse.__annotations__["links"]
            ):
                self_.links = kvargs["links"]
            else:
                self_.links = Links(**kvargs["links"])
        if "meta" in kvargs and kvargs["meta"] is not None:
            if (
                type(kvargs["meta"]).__name__
                == NLMetricsResponse.__annotations__["meta"]
            ):
                self_.meta = kvargs["meta"]
            else:
                self_.meta = NLMetricsResponseMeta(**kvargs["meta"])
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class NLMetricsResponseMeta:
    """

    Attributes
    ----------
    total: int
      The total number of metrics matching the current filter.
    """

    total: int = None

    def __init__(self_, **kvargs):
        if "total" in kvargs and kvargs["total"] is not None:
            self_.total = kvargs["total"]
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class NLResponses:
    """
    Provides info what was included in response for given query

    Attributes
    ----------
    hasChart: bool
      Chart was provided
    hasInsights: bool
      Narrative was provided
    hasMetadataApps: bool
      App list was provided
    hasMetadataDimensions: bool
      Dimensions list was provided
    hasMetadataMeasures: bool
      Measures list was provided
    hasSampleQueries: bool
      Sample questions was provided
    hasSuggestions: bool
      Suggestion questions was provided
    """

    hasChart: bool = None
    hasInsights: bool = None
    hasMetadataApps: bool = None
    hasMetadataDimensions: bool = None
    hasMetadataMeasures: bool = None
    hasSampleQueries: bool = None
    hasSuggestions: bool = None

    def __init__(self_, **kvargs):
        if "hasChart" in kvargs and kvargs["hasChart"] is not None:
            self_.hasChart = kvargs["hasChart"]
        if "hasInsights" in kvargs and kvargs["hasInsights"] is not None:
            self_.hasInsights = kvargs["hasInsights"]
        if "hasMetadataApps" in kvargs and kvargs["hasMetadataApps"] is not None:
            self_.hasMetadataApps = kvargs["hasMetadataApps"]
        if (
            "hasMetadataDimensions" in kvargs
            and kvargs["hasMetadataDimensions"] is not None
        ):
            self_.hasMetadataDimensions = kvargs["hasMetadataDimensions"]
        if (
            "hasMetadataMeasures" in kvargs
            and kvargs["hasMetadataMeasures"] is not None
        ):
            self_.hasMetadataMeasures = kvargs["hasMetadataMeasures"]
        if "hasSampleQueries" in kvargs and kvargs["hasSampleQueries"] is not None:
            self_.hasSampleQueries = kvargs["hasSampleQueries"]
        if "hasSuggestions" in kvargs and kvargs["hasSuggestions"] is not None:
            self_.hasSuggestions = kvargs["hasSuggestions"]
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class NarrativeResponse:
    """

    Attributes
    ----------
    text: str
    """

    text: str = None

    def __init__(self_, **kvargs):
        if "text" in kvargs and kvargs["text"] is not None:
            self_.text = kvargs["text"]
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class NluInfo:
    """

    Attributes
    ----------
    elements: list[NluInfoElements]
    """

    elements: list[NluInfoElements] = None

    def __init__(self_, **kvargs):
        if "elements" in kvargs and kvargs["elements"] is not None:
            if all(
                f"list[{type(e).__name__}]" == NluInfo.__annotations__["elements"]
                for e in kvargs["elements"]
            ):
                self_.elements = kvargs["elements"]
            else:
                self_.elements = [NluInfoElements(**e) for e in kvargs["elements"]]
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class NluInfoElements:
    """

    Attributes
    ----------
    entity: bool
    errorText: str
    filterFieldName: str
    filterText: str
    isFilter: bool
    text: str
    type: str
    typeName: str
    typeTranslated: str
    """

    entity: bool = None
    errorText: str = None
    filterFieldName: str = None
    filterText: str = None
    isFilter: bool = None
    text: str = None
    type: str = None
    typeName: str = None
    typeTranslated: str = None

    def __init__(self_, **kvargs):
        if "entity" in kvargs and kvargs["entity"] is not None:
            self_.entity = kvargs["entity"]
        if "errorText" in kvargs and kvargs["errorText"] is not None:
            self_.errorText = kvargs["errorText"]
        if "filterFieldName" in kvargs and kvargs["filterFieldName"] is not None:
            self_.filterFieldName = kvargs["filterFieldName"]
        if "filterText" in kvargs and kvargs["filterText"] is not None:
            self_.filterText = kvargs["filterText"]
        if "isFilter" in kvargs and kvargs["isFilter"] is not None:
            self_.isFilter = kvargs["isFilter"]
        if "text" in kvargs and kvargs["text"] is not None:
            self_.text = kvargs["text"]
        if "type" in kvargs and kvargs["type"] is not None:
            self_.type = kvargs["type"]
        if "typeName" in kvargs and kvargs["typeName"] is not None:
            self_.typeName = kvargs["typeName"]
        if "typeTranslated" in kvargs and kvargs["typeTranslated"] is not None:
            self_.typeTranslated = kvargs["typeTranslated"]
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class PartialNluInfo:
    """
    Contains break down of the asked question in the form of tokens with their classification.

    Attributes
    ----------
    fieldName: str
      Qlik sense application field selected for given token or phrase
    fieldValue: str
      Filter value found from query
    role: Literal["dimension", "measure", "date"]
      Role of the token or phrase from query
    text: str
      Matching token or phrase from query
    type: Literal["field", "filter", "master_dimension", "master_measure", "custom_analysis"]
      Type of token from query
    """

    fieldName: str = None
    fieldValue: str = None
    role: Literal["dimension", "measure", "date"] = None
    text: str = None
    type: Literal[
        "field", "filter", "master_dimension", "master_measure", "custom_analysis"
    ] = None

    def __init__(self_, **kvargs):
        if "fieldName" in kvargs and kvargs["fieldName"] is not None:
            self_.fieldName = kvargs["fieldName"]
        if "fieldValue" in kvargs and kvargs["fieldValue"] is not None:
            self_.fieldValue = kvargs["fieldValue"]
        if "role" in kvargs and kvargs["role"] is not None:
            self_.role = kvargs["role"]
        if "text" in kvargs and kvargs["text"] is not None:
            self_.text = kvargs["text"]
        if "type" in kvargs and kvargs["type"] is not None:
            self_.type = kvargs["type"]
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class QlikApp:
    """

    Attributes
    ----------
    id: str
    name: str
    """

    id: str = None
    name: str = None

    def __init__(self_, **kvargs):
        if "id" in kvargs and kvargs["id"] is not None:
            self_.id = kvargs["id"]
        if "name" in kvargs and kvargs["name"] is not None:
            self_.name = kvargs["name"]
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class QueryCreate:
    """

    Attributes
    ----------
    app: QlikApp
    clearEntityContext: bool
      Flag that clears the entity context.
    disableConversationContext: bool
      Flag that specifies either to enable converastion context.
    disableFollowups: bool
      The flag specifies whether to disable follow-up recommendations.
    disableNarrative: bool
      Flag that specifies whether the narratives should be generated for the user query or not.
    enableVisualizations: bool
      Flag that specifies whether visualization object should be provided or not.
    lang: str
      The language to assume when parsing, specified as an ISO-639-1 code.
      Defaults to 'en' (English).

    recommendationId: str
      property that contains the Id of the recommendation for which the response should be generated.
    text: str
      The sentence that will be parsed.
    visualizationTypes: list[str]
      Specify visualizationTypes for only which visualization object should be provided if enableVisualizations is set to true. For eg. ['linechart', 'barchart']
    """

    app: QlikApp = None
    clearEntityContext: bool = None
    disableConversationContext: bool = None
    disableFollowups: bool = None
    disableNarrative: bool = None
    enableVisualizations: bool = None
    lang: str = None
    recommendationId: str = None
    text: str = None
    visualizationTypes: list[str] = None

    def __init__(self_, **kvargs):
        if "app" in kvargs and kvargs["app"] is not None:
            if type(kvargs["app"]).__name__ == QueryCreate.__annotations__["app"]:
                self_.app = kvargs["app"]
            else:
                self_.app = QlikApp(**kvargs["app"])
        if "clearEntityContext" in kvargs and kvargs["clearEntityContext"] is not None:
            self_.clearEntityContext = kvargs["clearEntityContext"]
        if (
            "disableConversationContext" in kvargs
            and kvargs["disableConversationContext"] is not None
        ):
            self_.disableConversationContext = kvargs["disableConversationContext"]
        if "disableFollowups" in kvargs and kvargs["disableFollowups"] is not None:
            self_.disableFollowups = kvargs["disableFollowups"]
        if "disableNarrative" in kvargs and kvargs["disableNarrative"] is not None:
            self_.disableNarrative = kvargs["disableNarrative"]
        if (
            "enableVisualizations" in kvargs
            and kvargs["enableVisualizations"] is not None
        ):
            self_.enableVisualizations = kvargs["enableVisualizations"]
        if "lang" in kvargs and kvargs["lang"] is not None:
            self_.lang = kvargs["lang"]
        if "recommendationId" in kvargs and kvargs["recommendationId"] is not None:
            self_.recommendationId = kvargs["recommendationId"]
        if "text" in kvargs and kvargs["text"] is not None:
            self_.text = kvargs["text"]
        if "visualizationTypes" in kvargs and kvargs["visualizationTypes"] is not None:
            self_.visualizationTypes = kvargs["visualizationTypes"]
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class QueryCreated:
    """
    The attributes of sentences.

    Attributes
    ----------
    apps: list[QlikApp]
    conversationalResponse: list[QueryResponse]
      A list of conversational responses.
    nluInfo: NluInfo
    """

    apps: list[QlikApp] = None
    conversationalResponse: list[QueryResponse] = None
    nluInfo: NluInfo = None

    def __init__(self_, **kvargs):
        if "apps" in kvargs and kvargs["apps"] is not None:
            if all(
                f"list[{type(e).__name__}]" == QueryCreated.__annotations__["apps"]
                for e in kvargs["apps"]
            ):
                self_.apps = kvargs["apps"]
            else:
                self_.apps = [QlikApp(**e) for e in kvargs["apps"]]
        if (
            "conversationalResponse" in kvargs
            and kvargs["conversationalResponse"] is not None
        ):
            if all(
                f"list[{type(e).__name__}]"
                == QueryCreated.__annotations__["conversationalResponse"]
                for e in kvargs["conversationalResponse"]
            ):
                self_.conversationalResponse = kvargs["conversationalResponse"]
            else:
                self_.conversationalResponse = [
                    QueryResponse(**e) for e in kvargs["conversationalResponse"]
                ]
        if "nluInfo" in kvargs and kvargs["nluInfo"] is not None:
            if (
                type(kvargs["nluInfo"]).__name__
                == QueryCreated.__annotations__["nluInfo"]
            ):
                self_.nluInfo = kvargs["nluInfo"]
            else:
                self_.nluInfo = NluInfo(**kvargs["nluInfo"])
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class QueryNotCreated:
    """

    Attributes
    ----------
    apps: list[QlikApp]
    conversationalResponse: QueryResponse
    errors: list[Error]
    nluInfo: NluInfo
    """

    apps: list[QlikApp] = None
    conversationalResponse: QueryResponse = None
    errors: list[Error] = None
    nluInfo: NluInfo = None

    def __init__(self_, **kvargs):
        if "apps" in kvargs and kvargs["apps"] is not None:
            if all(
                f"list[{type(e).__name__}]" == QueryNotCreated.__annotations__["apps"]
                for e in kvargs["apps"]
            ):
                self_.apps = kvargs["apps"]
            else:
                self_.apps = [QlikApp(**e) for e in kvargs["apps"]]
        if (
            "conversationalResponse" in kvargs
            and kvargs["conversationalResponse"] is not None
        ):
            if (
                type(kvargs["conversationalResponse"]).__name__
                == QueryNotCreated.__annotations__["conversationalResponse"]
            ):
                self_.conversationalResponse = kvargs["conversationalResponse"]
            else:
                self_.conversationalResponse = QueryResponse(
                    **kvargs["conversationalResponse"]
                )
        if "errors" in kvargs and kvargs["errors"] is not None:
            if all(
                f"list[{type(e).__name__}]" == QueryNotCreated.__annotations__["errors"]
                for e in kvargs["errors"]
            ):
                self_.errors = kvargs["errors"]
            else:
                self_.errors = [Error(**e) for e in kvargs["errors"]]
        if "nluInfo" in kvargs and kvargs["nluInfo"] is not None:
            if (
                type(kvargs["nluInfo"]).__name__
                == QueryNotCreated.__annotations__["nluInfo"]
            ):
                self_.nluInfo = kvargs["nluInfo"]
            else:
                self_.nluInfo = NluInfo(**kvargs["nluInfo"])
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class QueryResponse:
    """

    Attributes
    ----------
    apps: list[QlikApp]
    contextInfo: str
      For contextual responses, this string contains a list of entities that are used to produce the response.
    drillDownURI: str
      The URL with the query injected to insight advisor of the app to which the query belongs.
    responses: list[ConversationalResponsePart]
    sentenceWithMatches: str
    """

    apps: list[QlikApp] = None
    contextInfo: str = None
    drillDownURI: str = None
    responses: list[ConversationalResponsePart] = None
    sentenceWithMatches: str = None

    def __init__(self_, **kvargs):
        if "apps" in kvargs and kvargs["apps"] is not None:
            if all(
                f"list[{type(e).__name__}]" == QueryResponse.__annotations__["apps"]
                for e in kvargs["apps"]
            ):
                self_.apps = kvargs["apps"]
            else:
                self_.apps = [QlikApp(**e) for e in kvargs["apps"]]
        if "contextInfo" in kvargs and kvargs["contextInfo"] is not None:
            self_.contextInfo = kvargs["contextInfo"]
        if "drillDownURI" in kvargs and kvargs["drillDownURI"] is not None:
            self_.drillDownURI = kvargs["drillDownURI"]
        if "responses" in kvargs and kvargs["responses"] is not None:
            if all(
                f"list[{type(e).__name__}]"
                == QueryResponse.__annotations__["responses"]
                for e in kvargs["responses"]
            ):
                self_.responses = kvargs["responses"]
            else:
                self_.responses = [
                    ConversationalResponsePart(**e) for e in kvargs["responses"]
                ]
        if (
            "sentenceWithMatches" in kvargs
            and kvargs["sentenceWithMatches"] is not None
        ):
            self_.sentenceWithMatches = kvargs["sentenceWithMatches"]
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class RecFeedback:
    """
    Any feedback from the user about a given recommendation

    Attributes
    ----------
    analysisType: str
    chartType: str
    comment: str
    recommendationAddedToHub: bool
    recommendationAddedToSheet: bool
    recommendationDisliked: bool
    recommendationLiked: bool
    """

    analysisType: str = None
    chartType: str = None
    comment: str = None
    recommendationAddedToHub: bool = None
    recommendationAddedToSheet: bool = None
    recommendationDisliked: bool = None
    recommendationLiked: bool = None

    def __init__(self_, **kvargs):
        if "analysisType" in kvargs and kvargs["analysisType"] is not None:
            self_.analysisType = kvargs["analysisType"]
        if "chartType" in kvargs and kvargs["chartType"] is not None:
            self_.chartType = kvargs["chartType"]
        if "comment" in kvargs and kvargs["comment"] is not None:
            self_.comment = kvargs["comment"]
        if (
            "recommendationAddedToHub" in kvargs
            and kvargs["recommendationAddedToHub"] is not None
        ):
            self_.recommendationAddedToHub = kvargs["recommendationAddedToHub"]
        if (
            "recommendationAddedToSheet" in kvargs
            and kvargs["recommendationAddedToSheet"] is not None
        ):
            self_.recommendationAddedToSheet = kvargs["recommendationAddedToSheet"]
        if (
            "recommendationDisliked" in kvargs
            and kvargs["recommendationDisliked"] is not None
        ):
            self_.recommendationDisliked = kvargs["recommendationDisliked"]
        if (
            "recommendationLiked" in kvargs
            and kvargs["recommendationLiked"] is not None
        ):
            self_.recommendationLiked = kvargs["recommendationLiked"]
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class RenderVisualization:
    """

    Attributes
    ----------
    data: object
      Data object should be used to render visualization
    language: str
    """

    data: object = None
    language: str = None

    def __init__(self_, **kvargs):
        if "data" in kvargs and kvargs["data"] is not None:
            self_.data = kvargs["data"]
        if "language" in kvargs and kvargs["language"] is not None:
            self_.language = kvargs["language"]
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


class Questions:
    def __init__(self, config: Config) -> None:
        self.config = config
        self.auth = Auth(config)

    def ask(self, data: QueryCreate, qlik_web_integration_id: str = None) -> any:
        """
        Returns the generated response for parsed chat queries, if no app was specified nor present in conversation context, suggests matching apps.

        Parameters
        ----------
        data: QueryCreate
        qlik_web_integration_id: str = None
          This header is only required for external clients or mashups for QCS, this value of this property should be the id of the web integration set up for the external client/mashup
        """
        if data is not None:
            try:
                data = asdict(data)
            except:
                data = data
        headers = {}
        if qlik_web_integration_id:
            headers["qlik-web-integration-id"] = qlik_web_integration_id
        response = self.auth.rest(
            path="/questions/actions/ask",
            method="POST",
            params={},
            data=data,
            headers=headers,
        )
        obj = QueryNotCreated(**response.json())
        obj.auth = self.auth
        return obj

    def filter(
        self,
        data: NLMetricsRequest,
        limit: int = 100,
        page: str = None,
        sort: Literal[
            "createdAt",
            "updatedAt",
            "+createdAt",
            "+updatedAt",
            "-createdAt",
            "-updatedAt",
        ] = "+createdAt",
    ) -> ListableResource[NLMetricRecord]:
        """
        Returns NL metrics based on provided app IDs the user has access to.

        Parameters
        ----------
        data: NLMetricsRequest
        limit: int = 100
          The preferred number of entries returned
        page: str = None
          A cursor pointing to the page of data to retrieve.
        sort: Literal["createdAt", "updatedAt", "+createdAt", "+updatedAt", "-createdAt", "-updatedAt"] = "+createdAt"
          A single field from the data model on which to sort the response. The '+' or '-' operator may be used to specify ascending or desending order.
        """
        if data is not None:
            try:
                data = asdict(data)
            except:
                data = data
        query_params = {}
        if limit is not None:
            query_params["limit"] = limit
        if page is not None:
            query_params["page"] = page
        if sort is not None:
            query_params["sort"] = sort
        response = self.auth.rest(
            path="/questions/actions/filter",
            method="POST",
            params=query_params,
            data=data,
        )
        return ListableResource(
            response=response.json(),
            cls=NLMetricRecord,
            auth=self.auth,
            path="/questions/actions/filter",
            query_params=query_params,
        )



# This is spectacularly generated code by spectacular based on
# Qlik Cloud Services APIs

from __future__ import annotations

from dataclasses import asdict, dataclass
from typing import Literal

from ..auth import Auth, Config
from ..listable import ListableResource


@dataclass
class CSPEntry:
    """

    Attributes
    ----------
    id: str
      The CSP entry's unique identifier.
    childSrc: bool
      Defines the valid sources for loading web workers and nested browsing contexts using elements such as frame and iFrame.
    connectSrc: bool
      Restricts the URLs that can be loaded using script interfaces.
    connectSrcWSS: bool
      Restricts the URLs that can be connected to websockets (all sources will be prefixed with 'wss://').
    createdDate: str
      The UTC timestamp when the CSP entry was created.
    description: str
      The reason for adding this origin to the Content Security Policy.
    fontSrc: bool
      Specifies valid sources for loading fonts.
    formAction: bool
      Allow forms to be submitted to the origin.
    frameAncestors: bool
      Specifies valid sources for embedding the resource using frame, iFrame, object, embed and applet.
    frameSrc: bool
      Specifies valid sources for loading nested browsing contexts using elements such as frame and iFrame.
    imgSrc: bool
      Specifies valid sources of images and favicons.
    mediaSrc: bool
      Specifies valid sources for loading media using the audio and video elements.
    modifiedDate: str
      The UTC timestamp when the CSP entry was last modified.
    name: str
      The name for this entry.
    objectSrc: bool
      Specifies valid sources for the object, embed, and applet elements.
    origin: str
      The origin that the CSP directives should be applied to.
    scriptSrc: bool
      Specifies valid sources for JavaScript.
    styleSrc: bool
      Specifies valid sources for stylesheets.
    workerSrc: bool
      Specifies valid sources for Worker, SharedWorker, or ServiceWorker scripts.
    """

    id: str = None
    childSrc: bool = None
    connectSrc: bool = None
    connectSrcWSS: bool = None
    createdDate: str = None
    description: str = None
    fontSrc: bool = None
    formAction: bool = None
    frameAncestors: bool = None
    frameSrc: bool = None
    imgSrc: bool = None
    mediaSrc: bool = None
    modifiedDate: str = None
    name: str = None
    objectSrc: bool = None
    origin: str = None
    scriptSrc: bool = None
    styleSrc: bool = None
    workerSrc: bool = None

    def __init__(self_, **kvargs):
        if "id" in kvargs and kvargs["id"] is not None:
            self_.id = kvargs["id"]
        if "childSrc" in kvargs and kvargs["childSrc"] is not None:
            self_.childSrc = kvargs["childSrc"]
        if "connectSrc" in kvargs and kvargs["connectSrc"] is not None:
            self_.connectSrc = kvargs["connectSrc"]
        if "connectSrcWSS" in kvargs and kvargs["connectSrcWSS"] is not None:
            self_.connectSrcWSS = kvargs["connectSrcWSS"]
        if "createdDate" in kvargs and kvargs["createdDate"] is not None:
            self_.createdDate = kvargs["createdDate"]
        if "description" in kvargs and kvargs["description"] is not None:
            self_.description = kvargs["description"]
        if "fontSrc" in kvargs and kvargs["fontSrc"] is not None:
            self_.fontSrc = kvargs["fontSrc"]
        if "formAction" in kvargs and kvargs["formAction"] is not None:
            self_.formAction = kvargs["formAction"]
        if "frameAncestors" in kvargs and kvargs["frameAncestors"] is not None:
            self_.frameAncestors = kvargs["frameAncestors"]
        if "frameSrc" in kvargs and kvargs["frameSrc"] is not None:
            self_.frameSrc = kvargs["frameSrc"]
        if "imgSrc" in kvargs and kvargs["imgSrc"] is not None:
            self_.imgSrc = kvargs["imgSrc"]
        if "mediaSrc" in kvargs and kvargs["mediaSrc"] is not None:
            self_.mediaSrc = kvargs["mediaSrc"]
        if "modifiedDate" in kvargs and kvargs["modifiedDate"] is not None:
            self_.modifiedDate = kvargs["modifiedDate"]
        if "name" in kvargs and kvargs["name"] is not None:
            self_.name = kvargs["name"]
        if "objectSrc" in kvargs and kvargs["objectSrc"] is not None:
            self_.objectSrc = kvargs["objectSrc"]
        if "origin" in kvargs and kvargs["origin"] is not None:
            self_.origin = kvargs["origin"]
        if "scriptSrc" in kvargs and kvargs["scriptSrc"] is not None:
            self_.scriptSrc = kvargs["scriptSrc"]
        if "styleSrc" in kvargs and kvargs["styleSrc"] is not None:
            self_.styleSrc = kvargs["styleSrc"]
        if "workerSrc" in kvargs and kvargs["workerSrc"] is not None:
            self_.workerSrc = kvargs["workerSrc"]
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)

    def delete(self) -> None:
        """
        Deletes a specific CSP entry

        Parameters
        ----------
        """
        self.auth.rest(
            path="/csp-origins/{id}".replace("{id}", self.id),
            method="DELETE",
            params={},
            data=None,
        )

    def set(self, data: CSPEntryContent) -> CSPEntry:
        """
        Updates a CSP entry

        Parameters
        ----------
        data: CSPEntryContent
        """
        if data is not None:
            try:
                data = asdict(data)
            except:
                data = data
        response = self.auth.rest(
            path="/csp-origins/{id}".replace("{id}", self.id),
            method="PUT",
            params={},
            data=data,
        )
        self.__init__(**response.json())
        return self


@dataclass
class CSPEntryContent:
    """

    Attributes
    ----------
    childSrc: bool
      Defines the valid sources for loading web workers and nested browsing contexts using elements such as frame and iFrame.
    connectSrc: bool
      Restricts the URLs that can be loaded using script interfaces.
    connectSrcWSS: bool
      Restricts the URLs that can be connected to websockets (all sources will be prefixed with 'wss://').
    createdDate: str
      The UTC timestamp when the CSP entry was created.
    description: str
      The reason for adding this origin to the Content Security Policy.
    fontSrc: bool
      Specifies valid sources for loading fonts.
    formAction: bool
      Allow forms to be submitted to the origin.
    frameAncestors: bool
      Specifies valid sources for embedding the resource using frame, iFrame, object, embed and applet.
    frameSrc: bool
      Specifies valid sources for loading nested browsing contexts using elements such as frame and iFrame.
    imgSrc: bool
      Specifies valid sources of images and favicons.
    mediaSrc: bool
      Specifies valid sources for loading media using the audio and video elements.
    modifiedDate: str
      The UTC timestamp when the CSP entry was last modified.
    name: str
      The name for this entry.
    objectSrc: bool
      Specifies valid sources for the object, embed, and applet elements.
    origin: str
      The origin that the CSP directives should be applied to.
    scriptSrc: bool
      Specifies valid sources for JavaScript.
    styleSrc: bool
      Specifies valid sources for stylesheets.
    workerSrc: bool
      Specifies valid sources for Worker, SharedWorker, or ServiceWorker scripts.
    """

    childSrc: bool = None
    connectSrc: bool = None
    connectSrcWSS: bool = None
    createdDate: str = None
    description: str = None
    fontSrc: bool = None
    formAction: bool = None
    frameAncestors: bool = None
    frameSrc: bool = None
    imgSrc: bool = None
    mediaSrc: bool = None
    modifiedDate: str = None
    name: str = None
    objectSrc: bool = None
    origin: str = None
    scriptSrc: bool = None
    styleSrc: bool = None
    workerSrc: bool = None

    def __init__(self_, **kvargs):
        if "childSrc" in kvargs and kvargs["childSrc"] is not None:
            self_.childSrc = kvargs["childSrc"]
        if "connectSrc" in kvargs and kvargs["connectSrc"] is not None:
            self_.connectSrc = kvargs["connectSrc"]
        if "connectSrcWSS" in kvargs and kvargs["connectSrcWSS"] is not None:
            self_.connectSrcWSS = kvargs["connectSrcWSS"]
        if "createdDate" in kvargs and kvargs["createdDate"] is not None:
            self_.createdDate = kvargs["createdDate"]
        if "description" in kvargs and kvargs["description"] is not None:
            self_.description = kvargs["description"]
        if "fontSrc" in kvargs and kvargs["fontSrc"] is not None:
            self_.fontSrc = kvargs["fontSrc"]
        if "formAction" in kvargs and kvargs["formAction"] is not None:
            self_.formAction = kvargs["formAction"]
        if "frameAncestors" in kvargs and kvargs["frameAncestors"] is not None:
            self_.frameAncestors = kvargs["frameAncestors"]
        if "frameSrc" in kvargs and kvargs["frameSrc"] is not None:
            self_.frameSrc = kvargs["frameSrc"]
        if "imgSrc" in kvargs and kvargs["imgSrc"] is not None:
            self_.imgSrc = kvargs["imgSrc"]
        if "mediaSrc" in kvargs and kvargs["mediaSrc"] is not None:
            self_.mediaSrc = kvargs["mediaSrc"]
        if "modifiedDate" in kvargs and kvargs["modifiedDate"] is not None:
            self_.modifiedDate = kvargs["modifiedDate"]
        if "name" in kvargs and kvargs["name"] is not None:
            self_.name = kvargs["name"]
        if "objectSrc" in kvargs and kvargs["objectSrc"] is not None:
            self_.objectSrc = kvargs["objectSrc"]
        if "origin" in kvargs and kvargs["origin"] is not None:
            self_.origin = kvargs["origin"]
        if "scriptSrc" in kvargs and kvargs["scriptSrc"] is not None:
            self_.scriptSrc = kvargs["scriptSrc"]
        if "styleSrc" in kvargs and kvargs["styleSrc"] is not None:
            self_.styleSrc = kvargs["styleSrc"]
        if "workerSrc" in kvargs and kvargs["workerSrc"] is not None:
            self_.workerSrc = kvargs["workerSrc"]
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class CSPEntryList:
    """

    Attributes
    ----------
    data: list[CSPEntry]
    links: CSPEntryListLinks
    """

    data: list[CSPEntry] = None
    links: CSPEntryListLinks = None

    def __init__(self_, **kvargs):
        if "data" in kvargs and kvargs["data"] is not None:
            if all(
                f"list[{type(e).__name__}]" == CSPEntryList.__annotations__["data"]
                for e in kvargs["data"]
            ):
                self_.data = kvargs["data"]
            else:
                self_.data = [CSPEntry(**e) for e in kvargs["data"]]
        if "links" in kvargs and kvargs["links"] is not None:
            if type(kvargs["links"]).__name__ == CSPEntryList.__annotations__["links"]:
                self_.links = kvargs["links"]
            else:
                self_.links = CSPEntryListLinks(**kvargs["links"])
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class CSPEntryListLinks:
    """

    Attributes
    ----------
    next: Link
    prev: Link
    self: Link
    """

    next: Link = None
    prev: Link = None
    self: Link = None

    def __init__(self_, **kvargs):
        if "next" in kvargs and kvargs["next"] is not None:
            if (
                type(kvargs["next"]).__name__
                == CSPEntryListLinks.__annotations__["next"]
            ):
                self_.next = kvargs["next"]
            else:
                self_.next = Link(**kvargs["next"])
        if "prev" in kvargs and kvargs["prev"] is not None:
            if (
                type(kvargs["prev"]).__name__
                == CSPEntryListLinks.__annotations__["prev"]
            ):
                self_.prev = kvargs["prev"]
            else:
                self_.prev = Link(**kvargs["prev"])
        if "self" in kvargs and kvargs["self"] is not None:
            if (
                type(kvargs["self"]).__name__
                == CSPEntryListLinks.__annotations__["self"]
            ):
                self_.self = kvargs["self"]
            else:
                self_.self = Link(**kvargs["self"])
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class CSPHeader:
    """

    Attributes
    ----------
    Content_Security_Policy: str
      The compiled CSP header.
    """

    Content_Security_Policy: str = None

    def __init__(self_, **kvargs):
        if (
            "Content-Security-Policy" in kvargs
            and kvargs["Content-Security-Policy"] is not None
        ):
            self_.Content_Security_Policy = kvargs["Content-Security-Policy"]
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class Link:
    """

    Attributes
    ----------
    href: str
      URL to a resource request.
    """

    href: str = None

    def __init__(self_, **kvargs):
        if "href" in kvargs and kvargs["href"] is not None:
            self_.href = kvargs["href"]
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


class CspOrigins:
    def __init__(self, config: Config) -> None:
        self.config = config
        self.auth = Auth(config)

    def generate_header(self) -> CSPHeader:
        """
        Retrieves the CSP header for a tenant

        Parameters
        ----------
        """
        response = self.auth.rest(
            path="/csp-origins/actions/generate-header",
            method="GET",
            params={},
            data=None,
        )
        obj = CSPHeader(**response.json())
        obj.auth = self.auth
        return obj

    def get(self, id: str) -> CSPEntry:
        """
        Returns details for a specific CSP entry

        Parameters
        ----------
        id: str
          The CSP entry's unique identifier.
        """
        response = self.auth.rest(
            path="/csp-origins/{id}".replace("{id}", id),
            method="GET",
            params={},
            data=None,
        )
        obj = CSPEntry(**response.json())
        obj.auth = self.auth
        return obj

    def get_csp_origins(
        self,
        childSrc: bool = None,
        connectSrc: bool = None,
        connectSrcWSS: bool = None,
        fontSrc: bool = None,
        formAction: bool = None,
        frameAncestors: bool = None,
        frameSrc: bool = None,
        imgSrc: bool = None,
        limit: float = 20,
        mediaSrc: bool = None,
        name: str = None,
        next: str = None,
        objectSrc: bool = None,
        origin: str = None,
        prev: str = None,
        scriptSrc: bool = None,
        sort: Literal[
            "name",
            "-name",
            "origin",
            "-origin",
            "createdDate",
            "-createdDate",
            "modifiedDate",
            "-modifiedDate",
        ] = None,
        styleSrc: bool = None,
        workerSrc: bool = None,
    ) -> ListableResource[CSPEntry]:
        """
        Retrieves all CSP entries for a tenant

        Parameters
        ----------
        childSrc: bool = None
          Filter resources by directive 'childSrc', true/false.
        connectSrc: bool = None
          Filter resources by directive 'connectSrc', true/false.
        connectSrcWSS: bool = None
          Filter resources by directive 'connectSrcWSS', true/false.
        fontSrc: bool = None
          Filter resources by directive 'fontSrc', true/false.
        formAction: bool = None
          Filter resources by directive 'formAction', true/false.
        frameAncestors: bool = None
          Filter resources by directive 'frameAncestors', true/false.
        frameSrc: bool = None
          Filter resources by directive 'frameSrc', true/false.
        imgSrc: bool = None
          Filter resources by directive 'imgSrc', true/false.
        limit: float = 20
          Maximum number of CSP-Origins to retrieve.
        mediaSrc: bool = None
          Filter resources by directive 'mediaSrc', true/false.
        name: str = None
          Filter resources by name (wildcard and case insensitive).
        next: str = None
          Cursor to the next page.
        objectSrc: bool = None
          Filter resources by directive 'objectSrc', true/false.
        origin: str = None
          Filter resources by origin (wildcard and case insensitive).
        prev: str = None
          Cursor to previous next page.
        scriptSrc: bool = None
          Filter resources by directive 'scriptSrc', true/false.
        sort: Literal["name", "-name", "origin", "-origin", "createdDate", "-createdDate", "modifiedDate", "-modifiedDate"] = None
          Field to sort by, prefix with -/+ to indicate order.
        styleSrc: bool = None
          Filter resources by directive 'styleSrc', true/false.
        workerSrc: bool = None
          Filter resources by directive 'workerSrc', true/false.
        """
        query_params = {}
        if childSrc is not None:
            query_params["childSrc"] = childSrc
        if connectSrc is not None:
            query_params["connectSrc"] = connectSrc
        if connectSrcWSS is not None:
            query_params["connectSrcWSS"] = connectSrcWSS
        if fontSrc is not None:
            query_params["fontSrc"] = fontSrc
        if formAction is not None:
            query_params["formAction"] = formAction
        if frameAncestors is not None:
            query_params["frameAncestors"] = frameAncestors
        if frameSrc is not None:
            query_params["frameSrc"] = frameSrc
        if imgSrc is not None:
            query_params["imgSrc"] = imgSrc
        if limit is not None:
            query_params["limit"] = limit
        if mediaSrc is not None:
            query_params["mediaSrc"] = mediaSrc
        if name is not None:
            query_params["name"] = name
        if next is not None:
            query_params["next"] = next
        if objectSrc is not None:
            query_params["objectSrc"] = objectSrc
        if origin is not None:
            query_params["origin"] = origin
        if prev is not None:
            query_params["prev"] = prev
        if scriptSrc is not None:
            query_params["scriptSrc"] = scriptSrc
        if sort is not None:
            query_params["sort"] = sort
        if styleSrc is not None:
            query_params["styleSrc"] = styleSrc
        if workerSrc is not None:
            query_params["workerSrc"] = workerSrc
        response = self.auth.rest(
            path="/csp-origins",
            method="GET",
            params=query_params,
            data=None,
        )
        return ListableResource(
            response=response.json(),
            cls=CSPEntry,
            auth=self.auth,
            path="/csp-origins",
            query_params=query_params,
        )

    def create(self, data: CSPEntryContent) -> CSPEntry:
        """
        Creates a new CSP entry

        Parameters
        ----------
        data: CSPEntryContent
        """
        if data is not None:
            try:
                data = asdict(data)
            except:
                data = data
        response = self.auth.rest(
            path="/csp-origins",
            method="POST",
            params={},
            data=data,
        )
        obj = CSPEntry(**response.json())
        obj.auth = self.auth
        return obj



from dataclasses import dataclass, field
from typing import List

from .auth_type import AuthType
from .errors import CustomException, CustomExceptionMsg


@dataclass
class Config:
    """Config object"""

    host: str
    """host adress of your tenant"""
    auth_type: AuthType
    """the authentication type"""
    client_id: str = None
    """oauth clientId"""
    client_secret: str = None
    """oauth clientSecret"""
    scope: List[str] = field(default_factory=list)
    """oauth scopes, always includes 'user_default'"""
    redirect_url: str = None
    """oauth redirect destination after successful authorization"""
    api_key: str = None
    """token representing a user in your tenant"""
    refresh_token: str = None
    """token required to fetch new access_token when current access_token expires"""

    def __post_init__(self):
        # Call validate after init
        self.validate()

    def validate(self):
        """
        Validate that the config is correct,
        raises error for incorrect config
        """
        if not self.host:
            raise CustomException(CustomExceptionMsg.EMPTY_HOST.value)
        if not self.auth_type:
            raise CustomException(
                CustomExceptionMsg.MISSING_CONFIG_PROPERTY.value + ": auth_type"
            )
        if self.auth_type == AuthType.APIKey:
            if not self.api_key:
                raise CustomException(
                    CustomExceptionMsg.MISSING_CONFIG_PROPERTY.value + ": api_key"
                )
            allowed_keys = {
                "host",
                "auth_type",
                "api_key",
                "scope",
            }  # scope has default
            for k in self.__dict__.keys():
                if k not in allowed_keys and self.__dict__[k] is not None:
                    raise CustomException(
                        CustomExceptionMsg.UNSUPPORTED_PROPERTY.value + ": " + k
                    )
        elif self.auth_type == AuthType.OAuth2:
            allowed_keys = {
                "host",
                "auth_type",
                "api_key",
                "client_id",
                "client_secret",
                "scope",
                "redirect_url",
                "refresh_token",
            }
            for k in self.__dict__.keys():
                if k not in allowed_keys and self.__dict__[k] is not None:
                    raise CustomException(
                        CustomExceptionMsg.UNSUPPORTED_PROPERTY.value + ": " + k
                    )
            # ensure scopes includes user_default
            if "user_default" not in self.scope:
                self.scope.append("user_default")
            if self.api_key is not None:
                raise CustomException(
                    CustomExceptionMsg.UNSUPPORTED_PROPERTY.value
                    + ": api_key cannot be used with oauth"
                )
            if not self.client_secret and not self.redirect_url:
                raise CustomException(
                    CustomExceptionMsg.MISSING_CONFIG_PROPERTY.value
                    + ": clientSecret or redirectUri"
                )

        else:
            raise CustomException(
                CustomExceptionMsg.UNSUPPORTED_AUTH_TYPE.value
                + ": "
                + str(self.auth_type)
            )



import io


def get_mime_type(buffer: io.BufferedReader):
    signature = ""
    if buffer:
        first_bytes = buffer.peek()
        signature = first_bytes.hex()[:8].upper()
    is_png = signature == "89504E47"
    is_jpg = (
        signature == "FFD8FFDB" or signature == "FFD8FFE0" or signature == "FFD8FFE1"
    )
    if is_png:
        return "image/png"
    if is_jpg:
        return "image/jpeg"
    return "application/octet-stream"
